"""
Internal module that handles common file operations.
"""

# SynDisco: Automated experiment creation and execution using only LLM agents
# Copyright (C) 2025 Dimitris Tsirmpas

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# You may contact the author at dim.tsirmpas@aueb.gr

import os
import json
import datetime
import logging

from typing import Any, Optional
from pathlib import Path


logger = logging.getLogger(Path(__file__).name)


def read_files_from_directory(directory: str | Path) -> list[str]:
    """Reads all files from a given directory.

    :param directory: the root directory from which to load files
        (NOT recursively!)
    :type directory: str | Path

    :raises ValueError: if the directory does not exist
    :return: Returns a list of parsed file content.
    :rtype: list[str | dict]
    """
    files_list = []

    # Check if directory exists
    if not os.path.isdir(directory):
        raise ValueError(f"Directory {directory} does not exist.")

    # Loop through all files in the directory
    for filename in os.listdir(directory):
        file_path = os.path.join(directory, filename)
        data = read_file(file_path)
        files_list.append(data)

    return files_list


def read_file(path: str | Path) -> str:
    """Return the contents of a file

    :param path: the path of the file
    :type path: str | Path
    :return: the file's contents
    :rtype: str
    """
    with open(path, "r", encoding="utf-8") as file:
        return file.read()


def read_json_file(path: str | Path) -> dict[str, Any]:
    """Read a JSON file
    :param path: the path of the file
    :type path: str | Path
    :return: the file's contents
    :rtype: dict[str, Any]
    """
    with open(path, "r", encoding="utf-8") as file:
        return json.load(file)


def dict_to_json(dictionary: dict[str, Any], output_path: str | Path) -> None:
    """
    Write a python dictionary to a file in JSON format.

    :param dictionary: The dictionary to be serialized
    :type dictionary: dict[str, Any]
    :param output_path: The output path to the file. If the parent directories
        do not exist, they will be created.
    :type output_path: str | Path
    """
    ensure_parent_directories_exist(output_path)

    with open(output_path, "w", encoding="utf8") as fout:
        json.dump(dictionary, fout, indent=4)


def ensure_parent_directories_exist(output_path: str | Path) -> None:
    """
    Create all parent directories if they do not exist.
    :param output_path: the path for which parent dirs will be generated
    :type output_path: str | Path
    """
    # Extract the directory path from the given output path
    directory = os.path.dirname(output_path)

    # Create all parent directories if they do not exist
    if directory:
        os.makedirs(directory, exist_ok=True)


def generate_datetime_filename(
    output_dir: Optional[str | Path] = None,
    timestamp_format: str = "%y-%m-%d-%H-%M",
    file_ending: str = "",
) -> Path:
    """
    Generate a filename based on the current date and time.

    :param output_dir: The path to the generated file, defaults to None
    :type output_dir: str | Path, optional
    :param timestamp_format: strftime format, defaults to "%y-%m-%d-%H-%M"
    :type timestamp_format: str, optional
    :param file_ending: The ending of the file (e.g '.json')
    :type file_ending: str
    :return: the full path for the generated file
    :rtype: str
    """
    datetime_name = (
        datetime.datetime.now().strftime(timestamp_format) + file_ending
    )
    path = ""
    if output_dir is None:
        path = Path(datetime_name)
    else:
        path = Path(os.path.join(output_dir, datetime_name))

    if not path.exists():
        path.parent.mkdir(parents=True, exist_ok=True)
        path.touch()

    return path



"""
Module handling the turn order of LLM participants in discussions.
"""

# SynDisco: Automated experiment creation and execution using only LLM agents
# Copyright (C) 2025 Dimitris Tsirmpas

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# You may contact the author at dim.tsirmpas@aueb.gr

import abc
import random
import warnings
import typing
from collections.abc import Iterable


class TurnManager(Iterable):
    """
    A class that handles which handles turns between users.
    """

    def __init__(self, names: Iterable[str] | None = None):
        """
        Construct a new TurnManager.

        :param names: The usernames of the participants.
            Can be left null if names are to be decided
            after this object's creation.
        :type config: dict[str, float], optional
        """
        self.names = [] if names is None else list(names)

    @typing.final
    def set_names(self, names: Iterable[str]) -> None:
        """
        Initialize the manager by providing the names of the users.

        :param names: the usernames of the participants
        :type names: Iterable[str]
        """
        self.names = list(names)

    @typing.final
    def next(self) -> str:
        """
        Get the username of the next speaker.

        :raises ValueError: if no names have been provided from the
            constructor, or from the TurnManager.set() method
        :return: the next speaker's username
        :rtype: str
        """
        if self.names == []:
            raise ValueError(
                "No usernames have been provided for the turn manager. "
                "Use self.initialize_names()"
            )
        return self._next_impl()

    def __iter__(self):
        return self

    def __next__(self):
        return self.next()

    @abc.abstractmethod
    def _next_impl(self) -> str:
        raise NotImplementedError("Abstract method called")


class RoundRobin(TurnManager):
    """
    A simple turn manager which gives priority to the next user in the queue.
    """

    def __init__(self, names: Iterable[str] | None = None):
        super().__init__(names)
        self.curr_turn = -1

    def _next_impl(self) -> str:
        self.curr_turn += 1
        new_speaker_index = self.curr_turn % len(self.names)
        return self.names[new_speaker_index]


class RandomWeighted(TurnManager):
    """
    Enable a participant to reply with a set probability, else randomly select
    another participant.
    """

    DEFAULT_RESPOND_PROBABILITY = 0.5

    def __init__(
        self, p_respond: float = -1, names: Iterable[str] | None = None
    ):
        super().__init__(names)

        if p_respond == -1:
            warnings.warn(
                "Warning: No p_respond set in RandomWeighted instance, "
                f"defaulting to {RandomWeighted.DEFAULT_RESPOND_PROBABILITY}"
            )
            self.chance_to_respond = RandomWeighted.DEFAULT_RESPOND_PROBABILITY
        else:
            self.chance_to_respond = p_respond
            assert (
                0 <= self.chance_to_respond <= 1
            ), f"p_respond must be between 0 and 1, but is {p_respond}"

        self.second_to_last_speaker = None
        self.last_speaker = None

    def _next_impl(self) -> str:
        # If first time asking for a speaker, return random speaker
        if self.second_to_last_speaker is None:
            next_speaker = self._select_other_random_speaker()
            self.last_speaker = next_speaker
            return next_speaker

        # Check if the last speaker will respond based on weighted coin flip
        if self._weighted_coin_flip():
            next_speaker = self.last_speaker
        else:
            next_speaker = self._select_other_random_speaker()

        # Update the speaker history
        self.second_to_last_speaker = self.last_speaker
        self.last_speaker = next_speaker

        assert next_speaker is not None
        return next_speaker

    def _weighted_coin_flip(self) -> bool:
        return self.chance_to_respond > random.uniform(0, 1)

    def _select_other_random_speaker(self) -> str:
        other_usernames = [
            username
            for username in self.names
            if username != self.last_speaker
        ]
        return random.choice(other_usernames)



"""
Module handling logging for LLM discussion and annotation tasks.
"""

# SynDisco: Automated experiment creation and execution using only LLM agents
# Copyright (C) 2025 Dimitris Tsirmpas

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# You may contact the author at dim.tsirmpas@aueb.gr

import time
import logging
import typing
import warnings
import functools
from pathlib import Path
from logging.handlers import TimedRotatingFileHandler

import coloredlogs


logger = logging.getLogger(Path(__file__).name)


def logging_setup(
    print_to_terminal: bool,
    write_to_file: bool,
    logs_dir: typing.Optional[str | Path] = None,
    level: str = "debug",
    use_colors: bool = True,
    log_warnings: bool = True,
) -> None:
    """
    Create the logger configuration.

    :param print_to_terminal: whether to print logs to the screen
    :type print_to_terminal: bool
    :param write_to_file: whether to write logs to a file.
        Needs logs_dir to be specified.
    :type write_to_file: bool
    :param logs_dir: the directory where the logs will be placed,
        defaults to None
    :type logs_dir: typing.Optional[str  |  Path], optional
    :param level: the logging level, defaults to logging.DEBUG
    :param use_colors: whether to color the output.
    :type use_colors: bool, defaults to True
    :param log_warnings: whether to log library warnings
    :type log_warnings: bool, defaults to True
    """
    if not print_to_terminal and logs_dir is None:
        warnings.warn(
            "Warning: Both screen-printing and file-printing has "
            "been disabled. No logs will be recorded for this session."
        )

    level = _str_to_log_level(level)  # type: ignore
    handlers = []
    if print_to_terminal:
        handlers.append(logging.StreamHandler())

    if write_to_file:
        if logs_dir is None:
            warnings.warn(
                "Warning: No logs directory provided ."
                "Disabling logging to file."
            )
        else:
            handlers.append(_get_file_handler(Path(logs_dir)))

    logging.basicConfig(
        handlers=handlers,
        level=level,
        format="%(asctime)s %(levelname)-8s %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )

    if use_colors:
        coloredlogs.install(level=level)

    logging.captureWarnings(log_warnings)


# https://stackoverflow.com/questions/1622943/timeit-versus-timing-decorator
def timing(f: typing.Callable) -> typing.Any:
    """
    Decorator which logs the execution time of a function.

    :param f: the function to be timed
    :type f: Function
    :return: the result of the function
    :rtype: _type_
    """

    @functools.wraps(f)
    def wrap(*args, **kw):
        ts = time.time()
        result = f(*args, **kw)
        te = time.time()
        exec_time_mins = (te - ts) / 60
        logger.debug(
            f"Procedure {f.__name__} executed in {exec_time_mins:2.4f} minutes"
        )
        return result

    return wrap


def _str_to_log_level(level_str: str):
    match level_str.lower().strip():
        case "debug":
            return logging.DEBUG
        case "info":
            return logging.INFO
        case "not_set":
            return logging.NOTSET
        case "warning":
            return logging.WARNING
        case "warn":
            return logging.WARNING
        case "error":
            return logging.ERROR
        case "critical":
            return logging.CRITICAL
        case _:
            logger.warning(
                f"Unrecognized log level {level_str}. Defaulting to NOT_SET"
            )
            return logging.NOTSET


def _get_file_handler(logs_dir: Path):
    logfile_path = logs_dir / "log"  # base filename, extension gets added
    file_handler = TimedRotatingFileHandler(
        filename=logfile_path,
        when="midnight",
        interval=1,
        backupCount=7,
        encoding="utf-8",
        utc=True,
    )
    file_handler.suffix = "%y-%m-%d.log"
    return file_handler



"""
Module responsible for exporting discussions and their annotations in CSV
format.
"""

# SynDisco: Automated experiment creation and execution using only LLM agents
# Copyright (C) 2025 Dimitris Tsirmpas

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# You may contact the author at dim.tsirmpas@aueb.gr

import os
import json
import hashlib
from pathlib import Path
from typing import Iterable

import pandas as pd


def import_discussions(conv_dir: Path) -> pd.DataFrame:
    """
    Import discussion output (logs) from JSON files in a directory and process
     it into a DataFrame.

    This function reads JSON files containing conversation data, processes the
     data to
    standardize columns, and adds derived attributes such as user traits and
     prompts.

    :param conv_dir: Directory containing JSON files with conversation data.
    :type conv_dir: str | Path
    :return: A DataFrame containing processed conversation data.
    :rtype: pd.DataFrame
    """
    df = _read_conversations(conv_dir)
    df = df.reset_index(drop=True)
    df = df.rename(columns={"id": "conv_id"})

    # Filter out non-persona information
    # assumes context and instructions are shared across participants
    df.user_prompts = df.user_prompts.apply(
        lambda user_prompts: [
            json.loads(user_prompt)["persona"] for user_prompt in user_prompts
        ]
    )

    # Select only current user prompt
    df["user_prompt"] = _select_user_prompt(df)

    # Merge moderator and user prompts
    df["is_moderator"] = _is_moderator(df.moderator, df.user)
    df.user_prompt = df.moderator_prompt.where(df.is_moderator, df.user_prompt)

    df["message_id"] = _generate_message_hash(df.conv_id, df.message)
    df["message_order"] = _add_message_order(df)

    traits_df = pd.concat(
        list(df.user_prompt.apply(_process_traits))
    ).reset_index(drop=True)
    del traits_df["username"]

    # Remove unused columns
    del df["user_prompts"]
    del df["user_prompt"]
    del df["users"]
    del df["moderator"]
    del df["moderator_prompt"]

    full_df = pd.concat([df, traits_df], axis=1)
    return full_df


def import_annotations(annot_dir: str | Path) -> pd.DataFrame:
    """
    Import annotation data from JSON files in a directory and process it
    into a DataFrame.

    This function reads JSON files containing annotation data, processes the
    data to standardize columns, and includes structured user traits.

    :param annot_dir: Directory containing JSON files with annotation data.
    :type annot_dir: str | Path
    :return: A DataFrame containing processed annotation data.
    :rtype: pd.DataFrame
    """
    annot_dir = Path(annot_dir)
    df = _read_annotations(annot_dir)
    df = df.reset_index(drop=True)
    df = _rename_annot_df_columns(df)

    # Generate unique message ID and message order
    df["message_id"] = _generate_message_hash(df.conv_id, df.message)
    df["message_order"] = _add_message_order(df)
    df = _group_all_but_one(df, "annot_personality_characteristics")
    return df


def _read_annotations(annot_dir: Path) -> pd.DataFrame:
    """
    Read annotation data from JSON files and convert it into a DataFrame.

    This function recursively reads all JSON files in the specified directory,
    extracts annotation data in raw form, and formats it into a DataFrame.

    :param annot_dir: Directory containing JSON files with annotation data.
    :type annot_dir: Path
    :return: A DataFrame containing raw annotation data.
    :rtype: pd.DataFrame
    """
    file_paths = _list_files_recursive(annot_dir)
    rows = []

    for file_path in file_paths:
        with open(file_path, "r", encoding="utf8") as fin:
            conv = json.load(fin)

        conv = pd.json_normalize(conv)
        conv = conv.explode("logs")
        conv["annotation_variant"] = os.path.basename(
            os.path.dirname(file_path)
        )
        conv["message"] = conv.logs.apply(lambda x: x[0])
        conv["annotation"] = conv.logs.apply(lambda x: x[1])

        del conv["logs"]
        rows.append(conv)

    full_df = pd.concat(rows)
    return full_df


def _rename_annot_df_columns(df):
    # Identify persona columns
    persona_prefix = "annotator_prompt.persona."
    rename_map = {
        col: "annot_" + col.replace(persona_prefix, "")
        for col in df.columns
        if col.startswith(persona_prefix)
    }
    # Apply renaming
    return df.rename(columns=rename_map)


def _read_conversations(conv_dir: Path) -> pd.DataFrame:
    """
    Read conversation data from JSON files and convert it into a DataFrame.

    This function recursively reads all JSON files in the specified directory,
    extracts conversation data in raw form, and formats it into a DataFrame.

    :param conv_dir: Directory containing JSON files with conversation data.
    :type conv_dir: str | Path
    :return: A DataFrame containing raw conversation data.
    :rtype: pd.DataFrame
    """
    if not conv_dir.is_dir():
        raise ValueError(
            f"{conv_dir} is not a directory or does not exist"
        ) from None

    file_paths = _list_files_recursive(conv_dir)

    if len(file_paths) == 0:
        raise ValueError(
            "No discussions found in directory ", conv_dir
        ) from None
    rows = []

    for file_path in file_paths:
        with open(file_path, "r", encoding="utf8") as fin:
            conv = json.load(fin)

        conv = pd.json_normalize(conv)
        conv = conv.explode("logs")
        conv["user"] = conv.logs.apply(lambda x: x["name"])
        conv["message"] = conv.logs.apply(lambda x: x["text"])
        conv["model"] = conv.logs.apply(lambda x: x["model"])
        del conv["logs"]
        rows.append(conv)

    full_df = pd.concat(rows)
    return full_df


def _is_moderator(moderator_name: pd.Series, username: pd.Series) -> pd.Series:
    """
    Determine if a user is the moderator.

    :param moderator_name: Series of moderator names.
    :type moderator_name: pd.Series
    :param username: Series of usernames.
    :type username: pd.Series
    :return: A Series indicating whether each user is the moderator.
    :rtype: pd.Series
    """
    return moderator_name == username


def _list_files_recursive(start_path: str | Path) -> list[str]:
    """
    Recursively list all files in a directory and its subdirectories.

    :param start_path: The starting directory path.
    :type start_path: str | Path
    :return: A list of file paths.
    :rtype: list[str]
    """
    all_files = []
    for root, _, files in os.walk(start_path):
        for file in files:
            all_files.append(os.path.join(root, file))
    return all_files


def _select_user_prompt(df):
    selected_user_prompts = []

    for _, row in df.iterrows():
        curr_username = row["user"]
        if curr_username == row["moderator"]:
            # Use moderator prompt
            selected_user_prompts.append(
                json.loads(row["moderator_prompt"])["persona"]
            )
            continue

        user_prompts = row["user_prompts"]
        matched_prompt = next(
            (p for p in user_prompts if p["username"] == curr_username),
            None,
        )
        if matched_prompt is None:
            raise ValueError(
                f"No matching prompt found for username: {curr_username}"
            )

        selected_user_prompts.append(matched_prompt)

    return selected_user_prompts


def _process_traits(user_prompt: dict) -> pd.DataFrame:
    """
    Process traits extracted from messages into a structured DataFrame.

    :param series: Series containing traits in dictionary format.
    :type series: pd.Series
    :return: A DataFrame with extracted traits as columns.
    :rtype: pd.DataFrame
    """
    return pd.DataFrame([user_prompt])


def _group_all_but_one(df: pd.DataFrame, to_list_col: str) -> pd.DataFrame:
    grouping_columns = [col for col in df.columns if col != to_list_col]
    aggregated_df = (
        df.groupby(grouping_columns, as_index=False)
        .agg({to_list_col: list})
        .reset_index(drop=True)
    )
    return aggregated_df


def _generate_message_hash(
    conv_ids: Iterable[str], messages: Iterable[str]
) -> list[str]:
    ls = []
    for conv_id, message in zip(conv_ids, messages):
        hashed_message = hashlib.md5(
            f"{conv_id}_{message}".encode()
        ).hexdigest()
        ls.append(hashed_message)
    return ls


def _add_message_order(df: pd.DataFrame) -> pd.Series:
    i = 1
    last_conv_id = -1
    last_message_id = -1
    numbers = []

    for _, row in df.iterrows():
        new_conv_id = row["conv_id"]
        new_message_id = row["message_id"]

        if new_conv_id != last_conv_id:
            last_conv_id = new_conv_id
            last_message_id = new_message_id
            i = 1
        elif new_message_id != last_message_id:
            i += 1
            last_message_id = new_message_id

        numbers.append(i)
    return pd.Series(numbers)



"""
Module containing wrappers for local LLMs loaded with various Python libraries.
"""

# SynDisco: Automated experiment creation and execution using only LLM agents
# Copyright (C) 2025 Dimitris Tsirmpas

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# You may contact the author at dim.tsirmpas@aueb.gr
import abc
import typing
import logging
from pathlib import Path

import transformers


logger = logging.getLogger(Path(__file__).name)


class BaseModel(abc.ABC):
    """
    Interface for all local LLM wrappers
    """

    def __init__(
        self,
        name: str,
        max_out_tokens: int,
        stop_list: list[str] | None = None,
    ):
        self.name = name
        self.max_out_tokens = max_out_tokens
        # avoid mutable default value problem
        self.stop_list = stop_list if stop_list is not None else []

    @typing.final
    def prompt(
        self,
        system_prompt: str,
        user_prompt: str,
    ) -> str:
        """Generate the model's response based on a prompt.

        :param system_prompt: The system prompt.
        :type system_prompt: str
        :param user_prompt: The user prompt.
        :type user_prompt: str
        :param stop_words: Strings to be removed after generation.
        :type stop_words: list[str]
        :return: the model's response
        :rtype: str
        """
        response = self._generate_response(system_prompt, user_prompt)
        # avoid model collapse attributed to certain strings
        for remove_word in self.stop_list:
            response = response.replace(remove_word, "")

        return response

    @typing.final
    def get_name(self) -> str:
        """
        Get the model's assigned pseudoname.

        :return: The name of the model.
        :rtype: str
        """
        return self.name

    @abc.abstractmethod
    def _generate_response(
        self,
        system_prompt: str,
        user_prompt: str,
    ) -> str:
        """Model-specific method which generates the LLM's response

        :param system_prompt: The system prompt.
        :type system_prompt: str
        :param user_prompt: The user prompt.
        :type user_prompt: str
        :return: The model's response
        :rtype: str
        """
        raise NotImplementedError("Abstract class call")


class TransformersModel(BaseModel):
    """
    HuggingFace Transformers model wrapper.
    """

    def __init__(
        self,
        model_path: str | Path,
        name: str,
        max_out_tokens: int,
        remove_string_list: list[str] | None = None,
    ):
        super().__init__(name, max_out_tokens, remove_string_list)

        self.model = transformers.AutoModelForCausalLM.from_pretrained(
            model_path, device_map="auto"
        )
        self.tokenizer = transformers.AutoTokenizer.from_pretrained(model_path)

        model_size = self.model.get_memory_footprint() / 2**20
        logger.info(f"Model memory footprint: {model_size:.2f} MB")

    def _generate_response(self, system_prompt: str, user_prompt: str) -> str:
        assert type(system_prompt) is str
        assert type(type(user_prompt) is str)
        # Construct proper message list for chat template
        messages = [
            {"role": "system", "content": system_prompt},
            {"role": "user", "content": user_prompt},
        ]

        # Prefer chat template if available
        if hasattr(self.tokenizer, "apply_chat_template"):
            prompt_text = self.tokenizer.apply_chat_template(
                messages,
                tokenize=False,
                add_generation_prompt=True,
            )
        else:
            logger.warning("Tokenizer has no chat template; falling back.")
            prompt_text = (
                f"System: {system_prompt}\nUser: {user_prompt}\nAssistant:"
            )

        inputs = self.tokenizer(prompt_text, return_tensors="pt").to(
            self.model.device
        )

        output_ids = self.model.generate(
            **inputs,
            max_new_tokens=self.max_out_tokens,
            do_sample=False,
            pad_token_id=self.tokenizer.eos_token_id,
        )

        # Remove the prompt portion → keep only generated part
        generated_ids = output_ids[0][inputs["input_ids"].shape[1]:]
        response = self.tokenizer.decode(
            generated_ids, skip_special_tokens=True
        )

        return response.strip()



"""
Module automating and managing batches of discussion/annotation tasks defined
in the syndisco.jobs module.
"""

# SynDisco: Automated experiment creation and execution using only LLM agents
# Copyright (C) 2025 Dimitris Tsirmpas

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# You may contact the author at dim.tsirmpas@aueb.gr

import time
import random
import logging
from pathlib import Path

from tqdm.auto import tqdm

from . import actors, turn_manager
from . import logging_util, _file_util
from . import jobs


logger = logging.getLogger(Path(__file__).name)


class DiscussionExperiment:
    """
    An experiment that creates, manages, and executes multiple synthetic
    discussions using LLM-based agents.
    """

    def __init__(
        self,
        users: list[actors.Actor],
        seed_opinions: list[list[str]] | None = None,
        moderator: actors.Actor | None = None,
        next_turn_manager: turn_manager.TurnManager | None = None,
        history_ctx_len: int = 3,
        num_turns: int = 10,
        num_active_users: int = 2,
        num_discussions: int = 5,
    ):
        """
        Initialize a synthetic discussion experiment.

        :param users: List of all possible participants (LLM agents).
        :type users: list[actors.Actor]
        :param seed_opinions: Hardcoded seed discussion
            segements to initiate synthetic discussions.
            One segment will be selected randomly for each new synthetic
            discussion and will be uttered by random synthetic participants.
            None if no seed opinions are to be provided.
        :type seed_opinions: list[list[str]], optional
        :param moderator: Optional moderator agent, or None to omit moderation.
        :type moderator: actors.Actor or None
        :param next_turn_manager: Strategy for selecting the next speaker.
            Defaults to round-robin if None.
        :type next_turn_manager: turn_manager.TurnManager or None
        :param history_ctx_len: Number of past comments visible as context.
        :type history_ctx_len: int
        :param num_turns: Number of user (non-moderator) turns per discussion.
        :type num_turns: int
        :param num_active_users: Number of active participants per discussion.
        :type num_active_users: int
        :param num_discussions: Total number of synthetic discussions to run.
        :type num_discussions: int
        """
        self.seed_opinions = (
            seed_opinions if seed_opinions is not None else [[]]
        )
        self.users = users
        self.moderator = moderator

        if next_turn_manager is None:
            logger.warning(
                "No TurnManager selected: Defaulting to round robin strategy."
            )
            self.next_turn_manager = turn_manager.RoundRobin()
        else:
            self.next_turn_manager = next_turn_manager

        self.history_ctx_len = history_ctx_len
        self.num_active_users = num_active_users
        self.num_discussions = num_discussions
        self.num_turns = num_turns

    def begin(
        self,
        discussions_output_dir: Path = Path("./output"),
        verbose: bool = True,
    ) -> None:
        """
        Generate and run all configured discussions.

        :param discussions_output_dir: Directory to write output JSON files.
        :type discussions_output_dir: Path
        :param verbose: Whether to print intermediate progress and outputs.
        :type verbose: bool
        """
        discussions = self._generate_discussions()
        self._run_all_discussions(discussions, discussions_output_dir, verbose)

    def _generate_discussions(self) -> list[jobs.Discussion]:
        """
        Internal helper to generate Discussion objects from configuration.

        :return: A list of configured Discussion objects.
        :rtype: list[jobs.Discussion]
        """
        experiments = []
        for _ in range(self.num_discussions):
            experiments.append(self._create_synthetic_discussion())
        return experiments

    def _create_synthetic_discussion(self):
        """
        Create and return a single randomized Discussion instance.

        :return: A synthetic Discussion object.
        :rtype: jobs.Discussion
        """
        rand_topic = random.choice(self.seed_opinions)
        rand_users = list(random.sample(self.users, k=self.num_active_users))

        return jobs.Discussion(
            users=rand_users,
            moderator=self.moderator,
            history_context_len=self.history_ctx_len,
            conv_len=self.num_turns,
            seed_opinions=rand_topic,
            next_turn_manager=self.next_turn_manager,
        )

    @logging_util.timing
    def _run_all_discussions(
        self,
        discussions: list[jobs.Discussion],
        output_dir: Path,
        verbose: bool,
    ) -> None:
        """
        Execute all generated discussions and write their outputs to disk.

        :param discussions: List of Discussion instances to run.
        :type discussions: list[jobs.Discussion]
        :param output_dir: Directory to save output JSON files.
        :type output_dir: Path
        :param verbose: Whether to print discussion progress.
        :type verbose: bool
        """
        output_dir.mkdir(parents=True, exist_ok=True)

        for i, discussion in tqdm(list(enumerate(discussions))):
            logging.info(
                f"Running experiment {i + 1}/{len(discussions) + 1}..."
            )
            self._run_single_discussion(
                discussion=discussion, output_dir=output_dir, verbose=verbose
            )

        logger.info("Finished synthetic discussion generation.")

    @logging_util.timing
    def _run_single_discussion(
        self, discussion: jobs.Discussion, output_dir: Path, verbose: bool
    ) -> None:
        """
        Run a single Discussion and store its results.

        :param discussion: The Discussion object to execute.
        :type discussion: jobs.Discussion
        :param output_dir: Directory to write the result file.
        :type output_dir: Path
        :param verbose: Whether to show detailed logging output.
        :type verbose: bool
        """
        try:
            logger.debug(f"Experiment parameters: {str(discussion)}")

            start_time = time.time()
            discussion.begin(verbose=verbose)
            output_path = _file_util.generate_datetime_filename(
                output_dir=output_dir, file_ending=".json"
            )
            logging.debug(
                f"Finished discussion in {(time.time() - start_time)} seconds."
            )

            discussion.to_json_file(output_path)
        except Exception:
            logger.exception("Experiment aborted due to error.")


class AnnotationExperiment:
    """
    An experiment that uses LLM annotators to label synthetic discussion logs.
    """

    def __init__(
        self,
        annotators: list[actors.Actor],
        history_ctx_len: int = 3,
        include_mod_comments: bool = True,
    ):
        """
        Initialize an annotation experiment using LLM-based annotators.

        :param annotators: List of annotator agents.
        :type annotators: list[actors.Actor]
        :param history_ctx_len: Number of previous comments visible to the
            annotator.
        :type history_ctx_len: int
        :param include_mod_comments: Whether to include moderator comments
            during annotation.
        :type include_mod_comments: bool
        """
        self.annotators = annotators
        self.history_ctx_len = history_ctx_len
        self.include_mod_comments = include_mod_comments

    def begin(
        self, discussions_dir: Path, output_dir: Path, verbose: bool = True
    ) -> None:
        """
        Start the annotation process over all discussion logs in a directory.

        :param discussions_dir: Directory containing discussion logs.
        :type discussions_dir: Path
        :param output_dir: Directory to write annotation outputs.
        :type output_dir: Path
        :param verbose: Whether to display annotation progress.
        :type verbose: bool, defaults to True
        """
        if not discussions_dir.is_dir():
            raise OSError(
                f"Discussions directory ({discussions_dir}) is not a directory"
            ) from None

        output_dir.mkdir(parents=True, exist_ok=True)

        annotation_tasks = self._generate_annotation_tasks(discussions_dir)
        self._run_all_annotations(annotation_tasks, output_dir, verbose)

    def _generate_annotation_tasks(
        self, discussions_dir: Path
    ) -> list[jobs.Annotation]:
        """
        Create annotation tasks by pairing each annotator with each discussion.

        :param discussions_dir: Path to discussion log files.
        :type discussions_dir: Path
        :return: List of Annotation tasks.
        :rtype: list[jobs.Annotation]
        """
        annotation_tasks = []
        for annotator in self.annotators:
            for discussion_path in discussions_dir.iterdir():
                annotation_task = self._create_annotation_task(
                    annotator, discussion_path
                )
                annotation_tasks.append(annotation_task)
        return annotation_tasks

    def _create_annotation_task(
        self, annotator: actors.Actor, conv_logs_path: Path
    ) -> jobs.Annotation:
        """
        Construct a single Annotation task.

        :param annotator: The LLM-based annotator.
        :type annotator: actors.Actor
        :param conv_logs_path: Path to the discussion log file.
        :type conv_logs_path: Path
        :return: Configured Annotation task.
        :rtype: jobs.Annotation
        """
        return jobs.Annotation(
            annotator=annotator,
            conv_logs_path=conv_logs_path,
            history_ctx_len=self.history_ctx_len,
            include_moderator_comments=self.include_mod_comments,
        )

    @logging_util.timing
    def _run_all_annotations(
        self,
        annotation_tasks: list[jobs.Annotation],
        output_dir: Path,
        verbose: bool = True,
    ) -> None:
        """
        Execute and store all annotation tasks.

        :param annotation_tasks: List of Annotation objects.
        :type annotation_tasks: list[jobs.Annotation]
        :param output_dir: Directory to save results.
        :type output_dir: Path
        :param verbose: Whether to log intermediate steps.
        :type verbose: bool, defaults to true
        """
        for annotation_task in tqdm(list(annotation_tasks)):
            self._run_single_annotation(annotation_task, output_dir, verbose)

        logger.info("Finished annotation generation.")

    @logging_util.timing
    def _run_single_annotation(
        self, annotation_task: jobs.Annotation, output_dir: Path, verbose: bool
    ) -> None:
        """
        Execute one annotation task and write its output.

        :param annotation_task: Single Annotation object to run.
        :type annotation_task: jobs.Annotation
        :param output_dir: Directory for output file.
        :type output_dir: Path
        :param verbose: Whether to show debug output.
        :type verbose: bool
        """
        try:
            logger.debug(f"Experiment parameters: {str(annotation_task)}")
            annotation_task.begin(verbose=verbose)
            output_path = _file_util.generate_datetime_filename(
                output_dir=output_dir, file_ending=".json"
            )
            annotation_task.to_json_file(output_path)
        except Exception:
            logger.exception("Annotation experiment aborted due to error.")



"""
Module handling the execution of LLM discussion and annotation tasks.
"""

# SynDisco: Automated experiment creation and execution using only LLM agents
# Copyright (C) 2025 Dimitris Tsirmpas

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# You may contact the author at dim.tsirmpas@aueb.gr

import collections
import datetime
import json
import logging
import uuid
import copy
import textwrap
import random
from pathlib import Path
from typing import Any, Optional

from tqdm.auto import tqdm

from . import actors, turn_manager
from . import _file_util


logger = logging.getLogger(Path(__file__).name)


# No superclass because the shared method names between the classes
# is coincidental


class Discussion:
    """
    A job conducting a discussion between different actors
    (:class:`actors.Actor`).
    """

    def __init__(
        self,
        next_turn_manager: turn_manager.TurnManager,
        users: list[actors.Actor],
        moderator: Optional[actors.Actor] = None,
        history_context_len: int = 5,
        conv_len: int = 5,
        seed_opinions: list[str] | None = None,
        seed_opinion_usernames: list[str] | None = None,
    ) -> None:
        """
        Construct the framework for a conversation to take place.

        :param turn_manager: an object handling the speaker priority of the
        participants
        :type turn_manager: turn_manager.TurnManager
        :param users: A list of discussion participants
        :type users: list[actors.Actor]
        :param moderator: An actor tasked with moderation if not None,
        can speak at any point in the conversation,
         defaults to None
        :type moderator: actors.Actor | None, optional
        :param history_context_len: How many prior messages are included
        to the LLMs prompt as context, defaults to 5
        :type history_context_len: int, optional
        :param conv_len: The total length of the conversation
        (how many times each actor will be prompted),
         defaults to 5
        :type conv_len: int, optional
        :param seed_opinions:
            The first hardcoded comments to start the discussion with.
            Will be inserted in the discussion, from top-to-bottom according
            to the ordering provided by the list.
        :type seed_opinion: list[str], optional
        :param seed_opinion_usernames:
            The usernames for each seed opinion.
            None if the usernames are to be selected randomly.
        :type seed_opinion_username:
            list[str], optional
        :raises ValueError: if the number of seed opinions and seed
        opinion users are different, or
        if the number of seed opinions exceeds history_context_len
        """
        users = copy.copy(users)
        self.username_user_map = {user.get_name(): user for user in users}

        self.next_turn_manager = next_turn_manager

        # used only during export, tags underlying models
        self.user_types = [
            type(user).__name__ for user in self.username_user_map
        ]

        self.moderator = moderator
        self.conv_len = conv_len

        # unique id for each conversation
        self.id = uuid.uuid4()

        # keep a limited context of the conversation to feed to the models
        self.ctx_len = history_context_len
        self.ctx_history = collections.deque(maxlen=history_context_len)
        self.conv_logs = []

        self.seed_opinions = seed_opinions or []
        self.seed_opinion_usernames = seed_opinion_usernames

    def begin(self, verbose: bool = True) -> None:
        self.next_turn_manager.set_names(list(self.username_user_map.keys()))

        if len(self.conv_logs) != 0:
            raise RuntimeError(
                "This conversation has already been concluded, "
                "create a new Discussion object."
            )

        self._add_seed_opinions(verbose)

        # begin main conversation
        for _ in tqdm(range(self.conv_len)):
            speaker_name = self.next_turn_manager.next()
            actor = self.username_user_map[speaker_name]
            res = actor.speak(list(self.ctx_history))

            if len(res.strip()) != 0:
                self._archive_response(actor, res, verbose)

                if self.moderator is not None:
                    res = self.moderator.speak(list(self.ctx_history))
                    self._archive_response(self.moderator, res, verbose)

    def to_dict(
        self, timestamp_format: str = "%y-%m-%d-%H-%M"
    ) -> dict[str, Any]:
        """
        Get a dictionary view of the data and metadata contained in the
        discussion.

        :param timestamp_format: the format for the conversation's creation
            time, defaults to "%y-%m-%d-%H-%M"
        :type timestamp_format: str, optional
        :return: a dict representing the conversation
        :rtype: dict[str, Any]
        """
        return {
            "id": str(self.id),
            "timestamp": datetime.datetime.now().strftime(timestamp_format),
            "users": [
                user.get_name() for user in self.username_user_map.values()
            ],
            "moderator": (
                self.moderator.get_name()
                if self.moderator is not None
                else None
            ),
            "user_prompts": [
                user.describe() for user in self.username_user_map.values()
            ],
            "moderator_prompt": (
                self.moderator.describe()
                if self.moderator is not None
                else None
            ),
            "ctx_length": self.ctx_len,
            "logs": self.conv_logs,
        }

    def to_json_file(self, output_path: str | Path) -> None:
        """
        Export the data and metadata of the conversation as a json file.

        :param output_path: the path for the exported file
        :type output_path: str
        """
        _file_util.dict_to_json(self.to_dict(), output_path)

    def _add_seed_opinions(self, verbose: bool) -> None:
        # Assign usernames if not provided
        if len(self.seed_opinions) > 0:
            usernames = self.seed_opinion_usernames
            if usernames is None:
                # sample without replacement
                if len(self.seed_opinions) > len(self.username_user_map):
                    raise ValueError(
                        "Not enough users to assign unique usernames "
                        "for seed opinions."
                    )
                usernames = random.sample(
                    list(self.username_user_map.keys()),
                    len(self.seed_opinions),
                )

            # insert seed opinions
            for username, comment in zip(usernames, self.seed_opinions):
                seed_user = actors.Actor(
                    model=None,  # type: ignore
                    persona=actors.Persona(username=username),
                    context="",
                    instructions="",
                    actor_type=actors.ActorType.USER,
                )
                if comment.strip() != "":
                    self._archive_response(seed_user, comment, verbose=verbose)

    def _archive_response(
        self, user: actors.Actor, comment: str, verbose: bool
    ) -> None:
        """
        Save the new comment to discussion output,
        to discussion history for other users to see, maybe print it on screen.

        :param user: The user who created the new comment.
        :type user: actors.LLMActor
        :param comment: The new comment.
        :type comment: str
        :param verbose: Whether to print the comment to stdout
        :type verbose: bool
        """
        self._log_comment(user, comment)
        self._add_comment_to_history(user, comment, verbose)

    def _log_comment(self, user: actors.Actor, comment: str) -> None:
        """
        Save new comment to the output history.

        :param user: The user who created the new comment
        :type user: actors.LLMActor
        :param comment: The new comment
        :type comment: str
        """
        model_name = (
            user.model.get_name() if user.model is not None else "hardcoded"
        )
        artifact = {
            "name": user.get_name(),
            "text": comment,
            "model": model_name,
        }
        self.conv_logs.append(artifact)

    def _add_comment_to_history(
        self, user: actors.Actor, comment: str, verbose: bool
    ) -> None:
        """
        Add new comment to the discussion history,
            so it can be shown to the other participants in the future.

        :param user: The user who created the new comment
        :type user: actors.LLMActor
        :param comment: The new comment
        :type comment: str
        :param verbose: Whether to print the comment to stdout
        :type verbose: bool
        """
        formatted_res = _format_chat_message(user.get_name(), comment)
        self.ctx_history.append(formatted_res)

        if verbose:
            print(formatted_res, "\n")

    def __str__(self) -> str:
        return json.dumps(self.to_dict(), indent=4)


class Annotation:
    """
    An annotation job modelled as a discussion between the system writing the
    logs of a finished discussion, and the LLM Annotator.
    """

    def __init__(
        self,
        annotator: actors.Actor,
        conv_logs_path: str | Path,
        include_moderator_comments: bool,
        history_ctx_len: int = 2,
    ):
        """
        Create an annotation job.
        The annotation is modelled as a conversation between the system and
        the annotator.

        :param annotator: The annotator
        :type annotator: actors.IActor
        :param conv_logs_path: The path to the file containing the
        discussion logs in JSON format
        :type conv_logs_path: str | Path
        :param include_moderator_comments: Whether to annotate moderator
        comments, and include them
        in conversational context when annotating user responses.
        :type include_moderator_comments: bool
        :param history_ctx_len: How many previous comments the annotator will
        remember, defaults to 4
        :type history_ctx_len: int, optional
        """
        self.annotator = annotator
        self.history_ctx_len = history_ctx_len
        self.include_moderator_comments = include_moderator_comments
        self.annotation_logs = []

        with open(conv_logs_path, "r", encoding="utf8") as fin:
            self.conv_data_dict = json.load(fin)

    def begin(self, verbose=True) -> None:
        """
        Begin the conversation-modelled annotation job.

        :param verbose: whether to print the results of the annotation to the
            console, defaults to True
        :type verbose: bool, optional
        """
        ctx_history = collections.deque(maxlen=self.history_ctx_len)

        for message_data in tqdm(self.conv_data_dict["logs"]):
            username = message_data["name"]
            message = message_data["text"]

            # do not include moderator comments in annotation ctx if told so
            if "moderator" in username:
                if not self.include_moderator_comments:
                    continue

            formatted_message = _format_chat_message(username, message)
            ctx_history.append(formatted_message)
            annotation = self.annotator.speak(list(ctx_history))
            self.annotation_logs.append((message, annotation))

            if verbose:
                print(textwrap.fill(formatted_message))
                print(annotation)

    def to_dict(
        self, timestamp_format: str = "%y-%m-%d-%H-%M"
    ) -> dict[str, Any]:
        """
        Get a dictionary view of the data and metadata contained in
        the annotation.

        :param timestamp_format: the format for the conversation's creation
            time, defaults to "%y-%m-%d-%H-%M"
        :type timestamp_format: str, optional
        :return: a dict representing the conversation
        :rtype: dict[str, Any]
        """
        return {
            "conv_id": str(self.conv_data_dict["id"]),
            "timestamp": datetime.datetime.now().strftime(timestamp_format),
            "annotator_model": self.annotator.model.get_name(),
            "annotator_prompt": self.annotator.describe(),
            "ctx_length": self.history_ctx_len,
            "logs": self.annotation_logs,
        }

    def to_json_file(self, output_path: str | Path) -> None:
        """
        Export the data and metadata of the conversation as a json file.

        :param output_path: the path for the exported file
        :type output_path: str
        """
        _file_util.dict_to_json(self.to_dict(), output_path)

    def __str__(self) -> str:
        return json.dumps(self.to_dict(), indent=4)


def _format_chat_message(username: str, message: str) -> str:
    """
    Create a prompt-friendly/console-friendly string representing a message
    made by a user.

    :param username: the name of the user who made the post
    :type username: str
    :param message: the message that was posted
    :type message: str
    :return: a formatted string containing both username and his message
    :rtype: str
    """
    if len(message.strip()) != 0:
        # append name of actor to his response
        # "user x posted" important for the model to not confuse it
        # with the instruction prompt
        wrapped_res = textwrap.fill(message, 70)
        formatted_res = f"User {username} posted:\n{wrapped_res}"
    else:
        formatted_res = ""

    return formatted_res



"""
Module defining LLM users in discussions and their characteristics.
"""

# SynDisco: Automated experiment creation and execution using only LLM agents
# Copyright (C) 2025 Dimitris Tsirmpas

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# You may contact the author at dim.tsirmpas@aueb.gr


import typing
import dataclasses
from pathlib import Path
import json
from enum import Enum, auto

from . import model
from . import _file_util


class ActorType(str, Enum):
    """
    The purpose of the LLMActor, used to determine proper prompt structure
    """

    USER = auto()
    ANNOTATOR = auto()


@dataclasses.dataclass
class Persona:
    """
    A dataclass holding information about the synthetic persona of a LLM actor.
    Includes name, Sociodemographic Background, personality
    and special instructions.
    """

    username: str = ""
    age: int = -1
    sex: str = ""
    sexual_orientation: str = ""
    demographic_group: str = ""
    current_employment: str = ""
    education_level: str = ""
    special_instructions: str = ""
    personality_characteristics: list[str] = dataclasses.field(
        default_factory=list
    )

    @staticmethod
    def from_json_file(file_path: Path) -> list:
        """
        Generate a list of personas from a properly formatted persona JSON
        file.

        :param file_path: the path to the JSON file containing the personas
        :type file_path: Path
        :return: a list of LlmPersona objects for each of the file entries
        :rtype: list[LlmPersona]
        """
        all_personas = _file_util.read_json_file(file_path)

        persona_objs = []
        for data_dict in all_personas:
            # code from https://stackoverflow.com/questions/68417319/initialize-python-dataclass-from-dictionary # noqa: E501
            field_set = {f.name for f in dataclasses.fields(Persona) if f.init}
            filtered_arg_dict = {
                k: v for k, v in data_dict.items() if k in field_set
            }
            persona_obj = Persona(**filtered_arg_dict)
            persona_objs.append(persona_obj)

        return persona_objs

    def to_dict(self):
        return dataclasses.asdict(self)

    def to_json_file(self, output_path: str) -> None:
        """
        Serialize the data to a .json file.

        :param output_path: The path of the new file
        :type output_path: str
        """
        _file_util.dict_to_json(self.to_dict(), output_path)

    def __str__(self):
        return json.dumps(self.to_dict())


class Actor:
    """
    An abstract class representing an actor which responds according to an
    underlying LLM instance.
    """

    def __init__(
        self,
        model: model.BaseModel,
        persona: Persona,
        context: str,
        instructions: str,
        actor_type: ActorType,
        stop_words: list[str] = [],
    ) -> None:
        """
        Create an Actor controlled by an LLM instance with a specific persona.

        :param model:
            A wrapper encapsulating a promptable LLM instance.
        :type model:
            model.BaseModel
        :param persona:
            The actor's persona.
        :type persona:
            persona.LLMPersona
        :param context:
            The context of the discussion.
        :type context:
            str
        :param instructions:
            The actor instructions for the discussion.
        :type instructions:
            str
        :param actor_type:
            Whether the actor is an annotator or participant.
        :type actor_type:
            ActorType
        """
        self.model = model
        self.persona = persona
        self.context = context
        self.instructions = instructions
        self.actor_type = actor_type

    def _system_prompt(self) -> str:
        prompt = {
            "context": self.context,
            "instructions": self.instructions,
            "type": self.actor_type,
            "persona": self.persona.to_dict(),
        }
        return json.dumps(prompt)

    def _message_prompt(self, history: list[str]) -> str:
        return _apply_template(self.actor_type, self.get_name(), history)

    @typing.final
    def speak(self, history: list[str]) -> str:
        """
        Prompt the actor to speak, given a history of previous messages
        in the conversation.

        :param history: A list of previous messages.
        :type history: list[str]
        :return: The actor's new message
        :rtype: str
        """
        system_prompt = self._system_prompt()
        message_prompt = self._message_prompt(history)
        response = self.model.prompt(system_prompt, message_prompt)
        return response

    def describe(self) -> str:
        """
        Get a description of the actor's internals.

        :return: A brief description of the actor
        :rtype: dict
        """
        return self._system_prompt()

    @typing.final
    def get_name(self) -> str:
        """
        Get the actor's assigned name within the conversation.

        :return: The name of the actor.
        :rtype: str
        """
        return self.persona.username


def _apply_template(
    actor_type: ActorType, username: str, history: list[str]
) -> str:

    if actor_type == ActorType.USER:
        json_input = {
            "role": "user",
            "content": f"{"\n".join(history)}\nUser {username} posted:",
        }
    elif actor_type == ActorType.ANNOTATOR:
        # LLMActor asks the model to respond as its username
        # by modifying this protected method, we instead prompt
        # it to write the annotation
        json_input = {
            "role": "user",
            "content": f"Conversation so far:\n{"\n".join(history)}\nOutput:",
        }
    return json.dumps(json_input)
